/*
** Module   : ProcedureManager.java
** Abstract : Implements access to procedure system handles and procedure
**            related attributes and methods. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130130 Added stubbed methods to implement the PUBLISH, SUBSCRIBE and UNSUBSCRIBE 
**                  language statements.
** 003 CA  20130131 Fixed problem determining THIS-PROCEDURE when the trigger invoked is defined 
**                  as ON ... PERSISTENT RUN.
** 004 CA  20130202 Update from character to TextOps.
** 005 CA  20130529 Some refactoring related to procedure deletion and added appserver support.
** 006 CA  20130918 Added runtime support for widget-pools: the pools need scope notifications,
**                  and this must be done by the ProcedureManager, to ensure proper access to the 
**                  THIS-PROCEDURE handle.
** 007 CA  20130925 Added support for named events.
** 008 CA  20131013 Implemented implicit deletion of static resources and delayed processing of
**                  finalizable/scopeable instances. Added support for the INSTANTIATING-PROCEDURE 
**                  attribute.
** 009 CA  20131208 Emit SEARCH-TARGET and SEARCH-SELF literals as integer.  Fixed runtime for 
**                  RUN SUPER and SUPER() calls, in search-target and search-self modes.
** 010 HC  20131222 Implemented ADM-DATA and UNIQUE-ID attribute.
**                  INSTANTIATING-PROCEDURE is now managed by ProcedureManager.
** 011 EVK 20131223 Made method isProcedure public.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.persist.StaticTempTable;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * Provides implementation for the following procedure handles:
 * <table>
 * <tr>
 *   <td>THIS-PROCEDURE</td>
 *   <td>{@link #thisProcedure()}</td>
 * </tr>
 * <tr>
 *   <td>SOURCE-PROCEDURE</td>
 *   <td>{@link #sourceProcedure()}</td>
 * </tr>
 * <tr>
 *   <td>TARGET-PROCEDURE</td>
 *   <td>{@link #targetProcedure()}</td>
 * </tr>
 * </table>
 * The following procedure-related attributes and methods are implemented:
 * <table>
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>SESSION:SUPER-PROCEDURES</td>
 *    <td>{@link #superProcedures()}</td>
 * </tr>
 * <tr>
 *    <td>SESSION:ADD-SUPER-PROCEDURE</td>
 *    <td>
 *       <ul>
 *          <li>{@link #addSuperProcedure(handle)}</li>
 *          <li>{@link #addSuperProcedure(handle, SearchMode)}</li>
 *          <li>{@link #addSuperProcedure(handle, int)}</li>
 *    </td>
 * </tr>
 * <tr>
 *    <td>SESSION:REMOVE-SUPER-PROCEDURE</td>
 *    <td>
 *       <ul>
 *          <li>{@link #removeSuperProcedure(handle)}</li>
 *       </ul>
 *    </td>
 * </tr>
 * <tr>
 *    <td>SESSION:FIRST-PROCEDURE</td>
 *    <td>{@link #firstProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>SESSION:LAST-PROCEDURE</td>
 *    <td>{@link #lastProcedure()}</td>
 * </tr>
 * </table>
 * <p>
 * The {@link #registerFunctionHandle(String, Resolvable)} is emitted at the
 * beginning of an external procedure, for any function defined 
 * <code>IN handle</code>.
 */
public class ProcedureManager
{
   /** Constant defining search mode: SEARCH-SELF */
   public static final int SEARCH_SELF = 1;
   
   /** Constant for socket search mode: SEARCH_TARGET */
   public static final int SEARCH_TARGET = 2;

   /** 
    * Defines search modes for the ADD-SUPER-PROCEDURE statement. Needed to
    * be able to identify the correct target in RUN SUPER and SUPER() cases.
    */
   public static enum SearchMode
   {
      SEARCH_SELF
      {
         @Override
         public String toString()
         {
            return "SEARCH-SELF";
         }
      },
      
      SEARCH_TARGET
      {
         @Override
         public String toString()
         {
            return "SEARCH-TARGET-PROCEDURE";
         }
      }
   };

   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer();

   /**
    * Register the handle expression for the given function.
    * 
    * @param    pname
    *           The function's legacy 4GL name.
    * @param    handleExpr
    *           The {@link Resolvable} instance used to determine its IN 
    *           handle target.
    */
   public static void registerFunctionHandle(String pname, 
                                             Resolvable handleExpr)
   {
      // always work with thisProcedure
      handle h = thisProcedure();
      Object referent = h.get();
      ProcedureData pp = locateProcedure(referent);
      pp.inHandleFunctions.put(pname, handleExpr);
   }
   
   /**
    * Get the THIS-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle thisProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.thisProcedures.peek();
      return new handle(locateProcedure(referent).phandle);
   }
   
   /**
    * Get the TARGET-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle targetProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.targetProcedures.peek();
      return new handle(locateProcedure(referent).phandle);
   }

   /**
    * Get the SOURCE-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle sourceProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.sourceProcedures.peek();

      return referent == null ? new handle()
                              : new handle(locateProcedure(referent).phandle);
   }
   
   /**
    * Get the SESSION:LAST-PROCEDURE handle (unknown if no persistent 
    * procedures are loaded).
    * 
    * @return   See above.
    */
   public static handle lastProcedure()
   {
      WorkArea wa = work.get();
      if (wa.persistentProcedures.isEmpty())
      {
         return new handle();
      }

      Object referent = wa.persistentProcedures.peekLast();
      return new handle(locateProcedure(referent).phandle);
   }
   
   /**
    * Get the SESSION:FIRST-PROCEDURE handle (unknown if no persistent 
    * procedures are loaded).
    * 
    * @return   See above.
    */
   public static handle firstProcedure()
   {
      WorkArea wa = work.get();
      if (wa.persistentProcedures.isEmpty())
      {
         return new handle();
      }

      Object referent = wa.persistentProcedures.peekFirst();     
      return new handle(locateProcedure(referent).phandle);
   }
   
   /**
    * Get the SESSION:SUPER-PROCEDURES attribute.
    * 
    * @return   See above.
    */
   public static character superProcedures()
   {
      WorkArea wa = work.get();
      return superProceduresImpl(wa.getSessionSuperProcedures().values());
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h)
   {
      return addSuperProcedure(h, (SearchMode) null);
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    searchMode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, int searchMode)
   {
      return addSuperProcedure(h, getSearchMode(searchMode));
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    searchMode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, integer searchMode)
   {
      if (searchMode.isUnknown())
      {
         return addSuperProcedure(h, (SearchMode) null);
      }
      else
      {
         return addSuperProcedure(h, searchMode.intValue());
      }
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the search-mode set as a string (one of the <code>SEARCH-TARGET</code>
    * or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, SearchMode mode)
   {
      return addSuperProcedureImpl(h, mode, null);
   }

   /**
    * Remove the given handle from the SESSION's list of super-procedures.
    *  
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical removeSuperProcedure(handle h)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }

      removeProcedureImpl(h, work.get().getSessionSuperProcedures().keySet());
   
      return new logical(true);
   }

   /**
    * Process the given resource by either setting its INSTANTIATING-PROCEDURE attribute 
    * immediately (in case of non-static resources) or add the static resource to the 
    * {@link WorkArea#pendingResources} set, to be processed when the next scope is entered.
    * 
    * @param    res
    *           The resource to be processed.
    * @param    dynamic
    *           Flag indicating if this is a dynamic or static resource.
    */
   public static void processResource(WrappedResource res, boolean dynamic)
   {
      // be a no-op in case of headless mode
      if (ErrorManager.isHeadless())
      {
         return;
      }
      
      WorkArea wa = work.get();
      
      boolean pending = !(res instanceof StaticTempTable);
      
      if (res instanceof DynamicResource && !dynamic && pending)
      {
         // all static resources are created as instance fields; thus, the actual processing is 
         // delayed, as the THIS-PROCEDURE is not available at the time of the call (i.e. we are
         // in an instance field initialization code, at the business logic class)
         wa.pendingResources.add(res);
      }
      else if (res instanceof HandleResource)
      {
         wa.pendingResources.remove(res);
         
         // all resources which don't have dynamic counterparts are processed immediately
         Object referent = _thisProcedure();
         ProcedureData pdata = locateProcedure(referent);

         // make sure the resource is registered as static
         if (!dynamic && !pending)
         {
            pdata.staticResources.peek().add(res);
         }

         ((HandleResource) res).setInstantiatingProcedure(new handle(pdata.phandle));
      }
   }
   
   /**
    * Get the current external program (an application logic instance) being processed for 
    * scopeable or finalizable notifications.
    * 
    * @return   See above
    */
   public static Object getProcessedProcedure()
   {
      WorkArea wa = work.get();
      
      return wa.processedProcedure;
   }
   
   /**
    * Check if we are executing finalizable or scopeable notifications.
    * 
    * @return   See above.
    */
   static boolean inNotifications()
   {
      return work.get().processedProcedure != null;
   }
   
   /**
    * Set the external procedure beeing processed for scopeable or finalizable notifications.
    * 
    * @param    referent
    *           The external procedure instance.
    */
   static void setProcessedProcedure(Object referent)
   {
      WorkArea wa = work.get();
      
      wa.processedProcedure = referent;
   }
   
   /**
    * Notify all subscribers (to the given event) that the event has occurred.
    * <p>
    * All processing is done with silent error mode implicitly enabled.
    *
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event. 
    *           Must not be <code>null</code>.
    * @param    modes
    *           The mode for each parameter, encoded in a string. When <code>null</code>, 
    *           the parameter modes will not be checked against the procedure definition.
    * @param    param
    *           The parameters passed to the subscriber. May be <code>null</code> to represent no
    *           parameters. When <code>null</code>, it means that no parameters were specified for
    *           this call.
    */
   static void publish(character event, handle publisher, String modes, Object[] param)
   {
      ErrorManager.silentErrorEnable();
      try
      {
         if (publisher.isUnknown())
         {
            final String msg = "Could not evaluate publisher handle expression while trying to " +
                               "execute PUBLISH statement";
            ErrorManager.recordOrThrowError(6472, msg, false);
            return;
         }
         if (!isProcedure(publisher) || !publisher._isValid())
         {
            final String msg = 
               "Invalid or inappropriate publisher handle value given to PUBLISH statement";
            ErrorManager.recordOrThrowError(9072, msg, false);
            return;
         }
         
         param = (param == null ? new Object[0] : param);
         
         // Progress does not validate the event
         
         Object rpub = publisher.get();
         
         // walk through all the procedures which are subscribed to this event and publisher and 
         // invoke their method
         WorkArea wa = work.get();
         for (ProcedureData pp : wa.pMap.values())
         {
            String pname = pp.pname.toStringMessage();

            // determine the subscriptions
            for (Subscription sub : pp.subscriptions)
            {
               // if the events match and the subscription is ANYWHERE or the publishers match,
               // then invoke the procedure
               if (sub.event.equals(event) &&
                  (sub.publisher == null || sub.publisher.equals(rpub)))
               {
                  // invoke the associated procedure. there are some rules here:
                  // - changes of INPUT-OUTPUT parameters are inherited from one call to another
                  // - OUTPUT parameters are always overridden
                  String iename = (sub.method.isUnknown() ? "" : sub.method.toStringMessage());
                  
                  // check if current procedure has a definition for target
                  String jname = SourceNameMapper.getMethodName(pname, iename, false);
                  if (jname == null)
                  {
                     final String msg = 
                        "PUBLISH statement failed. Procedure %s has no entry point for %s";
                     ErrorManager.recordOrThrowError(6480,
                                                     String.format(msg, pname, iename),
                                                     false);
                     continue;
                  }
                  
                  ExternalProgramWrapper extProg = new ExternalProgramWrapper(sub.subscriber);

                  // even if silent error mode is enabled by default, when the procedure is
                  // executed it will be disabled for the duration of the execution and re-enabled
                  // when the procedure ends. TransactionManager.push/popScope takes care of this.
                  ControlFlowOps.invokeIn(iename, new handle(extProg), param);
               }
            }
         }
      }
      finally
      {
         ErrorManager.silentErrorDisable();
      }
   }

   /**
    * Create a new subscription to the given event for a specified program. This program will 
    * receive a call-back when the event is published.
    *
    * @param    subscriber
    *           The program to notify when the specified event is published. Must not be 
    *           <code>null</code>.  The specified program must have a method that corresponds to 
    *           the given Progress 4GL internal procedure name (<code>method</code>).
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event. 
    *           If specified, the named event will only cause a notification if the event source 
    *           matches this handle. <code>null</code> represents the <code>ANYWHERE</code> option
    *           which means that the events will not be filtered by source.
    * @param    method
    *           The Progress 4GL internal procedure name which will be translated into a method 
    *           name and invoked on the <code>subscriber</code> instance. When <code>null</code>,
    *           it means that no method was specified for this call and the event name will be
    *           used as the method name.
    */
   static void subscribe(handle subscriber, character event, handle publisher, character method)
   {
      // validate subscriber
      if (subscriber.isUnknown())
      {
         final String msg = "Could not evaluate subscriber handle  expression while trying to " +
                            "execute SUBSCRIBE statement";
         ErrorManager.recordOrThrowError(6463, msg, false);
         return;
      }
      if (!isProcedure(subscriber) || !subscriber._isValid())
      {
         final String msg = "Invalid or inappropriate subscriber handle value given to " +
                            "SUBSCRIBE statement";
         ErrorManager.recordOrThrowError(6466, msg, false);
         return;
      }
      
      // validate publisher; if null, is OK, as it means ANYWHERE option
      if (publisher != null)
      {
         if (publisher.isUnknown())
         {
            final String msg = "Could not evaluate publisher handle  expression while trying " +
                               "to execute SUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6464, msg, false);
            return;
         }
         if (!isProcedure(publisher) || !publisher._isValid())
         {
            final String msg = "Invalid or inappropriate publisher handle value given to " +
                               "SUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6467, msg, false);
            return;
         }
      }
      // Progress does not validate event and method name
      
      Object rsub = subscriber.get();
      Object rpub = (publisher == null ? null : publisher.get());
      
      // 4GL doesn't raise an error if a subscription already exists, but doesn't replace it 
      // either
      ProcedureData pp = locateProcedure(rsub);
      boolean found = false;
      for (Subscription sub : pp.subscriptions)
      {
         if (sub.readOnly)
         {
            // read-only subscriptions are not checked.
            continue;
         }

         // get an exact match
         if (event.equals(sub.event) && 
             ((rpub == null && sub.publisher == null) || 
              (rpub != null && sub.publisher != null && rpub.equals(sub.publisher))))
         {
            found = true;
            break;
         }
      }

      // create only if there isn't an existing one
      if (!found)
      {
         Subscription sub = new Subscription(rsub, event, rpub, method);
         pp.subscriptions.add(sub);
      }
   }

   /**
    * Create a new subscription to the given event for any program. This program will receive a 
    * call-back when the event is published.
    *
    * @param    subscriber
    *           The program to notify when the specified event is published. Must not be 
    *           <code>null</code>.  The specified program must have a method that corresponds to 
    *           the given Progress 4GL internal procedure name (<code>method</code>).
    * @param    event
    *           The name of the event which is being published.
    * @param    method
    *           The Progress 4GL internal procedure name which will be translated into a method 
    *           name and invoked on the <code>subscriber</code> instance. When <code>null</code>,
    *           it means that no method was specified for this call and the event name will be
    *           used as the method name.
    */
   static void subscribeAnywhere(handle subscriber, character event, character method)
   {
      subscribe(subscriber, event, null, method);
   }
   
   /**
    * Delete a subscription to the given event for the specified program. This method is executed
    * in a {@link ErrorManager#silentErrorEnable()}/{@link ErrorManager#silentErrorDisable()}
    * bracket, as the <code>UNSUBSCRIBE</code> statement is executed with a default NO-ERROR 
    * clause.
    *
    * @param    subscriber
    *           The program whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    event
    *           The name of the event which is being published. If this is <code>null</code>, then
    *           subscriptions to all events will be deleted.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source. When <code>null</code>, it means
    *           that the publisher was not set.
    */
   static void unsubscribe(handle subscriber, character event, handle publisher)
   {
      ErrorManager.silentErrorEnable();

      try
      {
         // validate subscriber
         if (subscriber.isUnknown())
         {
            final String msg = "Could not evaluate subscriber handle  expression while trying " +
                               "to execute UNSUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6475, msg, false);
            return;
         }
         if (!isProcedure(subscriber) || !subscriber._isValid())
         {
            final String msg = "Invalid or inappropriate subscriber handle value given to " +
                               "UNSUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6477, msg, false);
            return;
         }
         
         // validate publisher; if null, is OK, as it means ALL option
         if (publisher != null)
         {
            if (publisher.isUnknown())
            {
               final String msg = "Could not evaluate publisher handle  expression while " +
                                  "trying to execute UNSUBSCRIBE statement";
               ErrorManager.recordOrThrowError(6476, msg, false);
               return;
            }
            if (!isProcedure(publisher) || !publisher._isValid())
            {
               final String msg = "Invalid or inappropriate publisher handle value given to " +
                                  "UNSUBSCRIBE statement";
               ErrorManager.recordOrThrowError(6478, msg, false);
               return;
            }
         }
         
         // Progress does not validate the event
         
         Object rsub = subscriber.get();
         Object rpub = (publisher == null ? null : publisher.get());
         
         ProcedureData pp = locateProcedure(rsub);
         boolean found = false;
         Iterator<Subscription> itr = pp.subscriptions.iterator();
         while (itr.hasNext())
         {
            Subscription sub = itr.next();

            // a special case: if a publisher is specified and ALL clause is used, than all 
            // ANYWERE subscriptions are marked read-only, for the event.
            if (rpub != null && event == null && sub.publisher == null)
            {
               sub.readOnly = true;
            }
            
            if (sub.readOnly)
            {
               // read-only subscriptions can't be removed
               continue;
            }
            
            // if removes all subscriptions considering:
            // - if the passed event is null, then ALL clause is in effect, else event names must
            // match
            // - the publisher is not matched as stated in the documentation. the correct approach
            // is that if the publisher is missing, then ANY publisher is matched
            if ((event == null || event.equals(sub.event)) &&
                ((rpub == null || rpub.equals(sub.publisher))))
            {
               // remove this subscription
               itr.remove();
               found = true;
            }
         }

         // if a specific event is used and no subscriptions found, then error
         if (!found && event != null)
         {
            final String msg = "UNSUBSCRIBE failed.  No subscriptions for %s";
            String sevt = (event.isUnknown() ? "" : event.toStringMessage());
            ErrorManager.recordOrThrowError(6479, String.format(msg, sevt), false);
            return;
         }
      }
      finally
      {
         ErrorManager.silentErrorDisable();
      }
   }

   /**
    * Delete a subscription to ALL events for the specified program.
    *
    * @param    subscriber
    *           The program whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source.  When <code>null</code>, it means
    *           that the publisher was not set.
    */
   static void unsubscribeAll(handle subscriber, handle publisher)
   {
      unsubscribe(subscriber, null, publisher);
   }

   /**
    * Ensure all the context-local data has initialized (to be able to listen
    * for scope notifications).
    */
   static void load()
   {
      // needed to ensure the variable is loaded, so it listens for scope
      // notifications
      work.get();
   }
   
   /**
    * Check if the handle's referent is a procedure or not. This means that the value must be set 
    * and must be an instance of {@link ExternalProgramWrapper}. The enclosed referent may or may
    * not be registered with the {@link ProcedureManager}.
    *
    * @param    h
    *           handle object to check.
    *  
    * @return   See above.
    */
   public static boolean isProcedure(handle h)
   {
      return h != null && !h.isUnknown() && (h.getResource() instanceof ExternalProgramWrapper);
   }

   /**
    * Add the <code>h</code> handle as a persistent procedure. Used by 
    * {@link ControlFlowOps#invokePersistent(character, Object...)} and
    * {@link ControlFlowOps#invokePersistentSet(String, handle, Object...)}.
    * 
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    */
   static void addPersistentProcedure(handle h)
   {
      WorkArea wa = work.get();
      
      removeProcedureImpl(h, wa.persistentProcedures);
      ProcedureData pp = locateProcedure(h.get());
      if (pp == null)
      {
         pp = new ProcedureData(h, null);
         wa.pMap.put(pp.phandle.get(), pp);
      }

      wa.persistentProcedures.add(pp.phandle.get());
   }

   /**
    * Remove the given handle from the list of persistent procedures. Used by 
    * {@link ControlFlowOps#invokePersistent(character, Object...)} and
    * {@link ControlFlowOps#invokePersistentSet(String, handle, Object...)} in
    * cases when errors were encountered while invoking the procedure.
    * 
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    */
   static void removePersistentProcedure(handle h)
   {
      removeProcedureImpl(h, work.get().persistentProcedures);
   }

   /**
    * Get the FILE-NAME attribute of the given procedure handle, passed as
    * the external procedure instance. 
    * Used by {@link ExternalProgramWrapper#getFileName()} and other internal
    * APIs.
    * 
    * @param    referent
    *           The referent for which the name is needed.
    *           
    * @return   The FILE-NAME attribute.
    */
   static character getFileName(Object referent)
   {
      String jname = referent.getClass().getName();
      String pname = SourceNameMapper.convertJavaProg(jname);
      return new character(pname);
   }

   /**
    * Check if the given referent has been (or is still) registered with the
    * {@link ProcedureManager} in the {@link WorkArea#pMap procedure} map.
    * 
    * @param     referent
    *            The referent to be checked if is still loaded.
    * 
    * @return    <code>true</code> if the referent was found.
    */
   static boolean hasReferent(Object referent)
   {
      return work.get().pMap.containsKey(referent);
   }
   
   /**
    * Delete the given procedure handle (i.e. execute the DELETE OBJECT or DELETE PROCEDURE 
    * handle).
    * 
    * @param    referent
    *           The persistent procedure instance to be deleted.
    * 
    * @return   <code>true</code> if this procedure handle can be deleted.
    */
   static boolean delete(Object referent)
   {
      // handle is always valid
      WorkArea wa = work.get();

      ProcedureData pdata = locateProcedure(referent);
      
      // if there is an active call on the stack for the specified referent, do not delete
      if (wa.thisProcedures.contains(referent) ||
          wa.targetProcedures.contains(referent) ||
          wa.sourceProcedures.contains(referent))
      {
         return false;
      }
      
      wa.persistentProcedures.remove(referent);
      wa.sessionSuperProcs.remove(referent);
      
      for (ProcedureData pp : wa.pMap.values())
      {
         pp.superProcedures.remove(referent);
      }

      wa.cleanup(referent);
      
      try
      {
         wa.processedProcedure = referent;
         
         // cleanup after the procedure
         for (Finalizable target : pdata.finalizables)
         {
            target.deleted();
         }
         
         for (Scopeable target : pdata.scopeables)
         {
            target.scopeDeleted();
         }

         // as the procedure no longer exists, delete the static resources
         deleteResources(pdata.staticResources.pop());
      }
      finally
      {
         wa.processedProcedure = null;
      }

      // delete the widget pools for this external program
      WidgetPool.deleteWidgetPools(referent);
      
      return true;
   }
   
   /**
    * Push the new SOURCE-PROCEDURE onto the {@link WorkArea#sourceProcedures}
    * stack (this pushes the current THIS-PROCEDURE, as this is invoked before
    * a function or procedure call).
    */
   static void pushSourceProcedure()
   {
      handle h = thisProcedure();
      work.get().sourceProcedures.push(h.get());
   }

   /**
    * Pop the SOURCE-PROCEDURE from {@link WorkArea#sourceProcedures} stack.
    */
   static void popSourceProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.sourceProcedures.pop();
      wa.cleanup(referent);
   }

   /**
    * Push the given handle as THIS-PROCEDURE.
    * 
    * @param    h
    *           A handle containing the new THIS-PROCEDURE.
    */
   static void pushThisProcedure(handle h)
   {
      work.get().thisProcedures.push(h.get());
   }

   /**
    * Pop the latest THIS-PROCEDURE from the stack.
    */
   static void popThisProcedure()
   {
      work.get().thisProcedures.pop();
   }

   /**
    * Push the given handle as the new TARGET-PROCEDURE on the 
    * {@link WorkArea#targetProcedures} stack.
    * 
    * @param    h
    *           The new TARGET-PROCEDURE.
    */
   static void pushTargetProcedure(handle h)
   {
      WorkArea wa = work.get();
      Object referent = h.get();
      wa.targetProcedures.push(referent);
   }

   /**
    * Pop the TARGET-PROCEDURE from the stack.
    */
   static void popTargetProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.targetProcedures.pop();
      wa.cleanup(referent);
   }

   /**
    * Remove the given handle from the available super-procedures, for target-procedure.
    * 
    * @param    h
    *           The used super-procedure.
    * @param    superCall
    *           Flag indicating if this is a RUN SUPER or SUPER() call.
    */
   static void pushSuperProcedure(handle h, boolean superCall)
   {
      WorkArea wa = work.get();
      Object referent = wa.targetProcedures.peek();
      ProcedureData ptarget = locateProcedure(referent);
      
      Deque<Deque<Object>> availProcs = ptarget.availSuperProcedures;
      if (!superCall)
      {
         availProcs.push(new ArrayDeque<>(ptarget.superProcedures.keySet()));
      }
      
      // remove this from the available procedures
      availProcs.peek().remove(h.get());
   }

   /**
    * Pop the last super-procedure data for, if this is not a RUN SUPER or SUPER() call.
    * 
    * @param    superCall
    *           Flag indicating if this is a RUN SUPER or SUPER() call.
    */
   static void popSuperProcedure(boolean superCall)
   {
      WorkArea wa = work.get();
      Object referent = wa.targetProcedures.peek();
      ProcedureData ptarget = locateProcedure(referent);

      Deque<Deque<Object>> availProcs = ptarget.availSuperProcedures;
      if (!superCall)
      {
         availProcs.pop();
      }
   }

   /**
    * Using the name of a function defined IN handle and a procedure handle 
    * to which this function definition belongs, resolve the target handle,
    * to be able to invoke it.
    * 
    * @param    phandle
    *           The procedure handle where the function was defined.
    * @param    fname
    *           The function's legacy name.
    *           
    * @return   The procedure handle where to search for the function's body.
    */
   static handle getHandleForFunction(handle phandle, String fname)
   {
      Object referent = phandle.get();
      ProcedureData pp = locateProcedure(referent);
   
      handle h = (handle) pp.inHandleFunctions.get(fname).resolve();
      return h == null ? new handle() : h;
   }

   /**
    * For procedures or functions defined IN SUPER, this computes the target procedure handle 
    * where the procedure's or function's body resides.
    * <p>
    * In case when no super procedure is found, it will throw proper error messages and return
    * <code>null</code>. 
    * 
    * @param    iename
    *           Internal-entry's name (this may be a procedure or function).
    * @param    function
    *           <code>true</code> if this is a function.
    *           
    * @return   A procedure handle where to search for the internal-entry's body or
    *           <code>null</code> of none found.
    */
   static handle getSuperHandle(String iename, boolean function)
   {
      WorkArea wa = work.get();
      
      handle hthis   = thisProcedure();
      handle htarget = targetProcedure();
      ProcedureData ppThis   = locateProcedure(hthis.get());
      ProcedureData ppTarget = locateProcedure(htarget.get());

      boolean searchSelf = ppThis.isSearchSelf();
      handle psuper = null;
      
      if (searchSelf || !hthis.equals(htarget))
      {
         // search-target mode is searched only if this-proc is different than target-proc

         Collection<Object>[] search = new Collection[2];
         search[0] = null;
         search[1] = wa.sessionSuperProcs.keySet();
         
         // walk the list of this-procedure's
         if (searchSelf)
         {
            search[0] = ppThis.superProcedures.keySet();
         }
         else
         {
            search[0] = ppTarget.availSuperProcedures.peek();
         }
         
         for (int i = 0; i < search.length && psuper == null; i++)
         {
            if (search[i] == null)
            {
               continue;
            }
            
            // search is done backwards (last set super-procedure is searched   first)
            Deque<Object> queue = new ArrayDeque<>(search[i]);
            while (!queue.isEmpty())
            {
               Object referent = queue.removeLast();
               ProcedureData pp = locateProcedure(referent);
   
               if (!searchSelf)
               {
                  // if in search-target mode, remove any touched procedure
                  search[0].remove(referent);
               }
   
               if (pp == null || !pp.phandle._isValid())
               {
                  // delete or invalid procedures are ignored
                  continue;
               }
   
               character pname = pp.pname;
               String spname = pname.toStringMessage();
               String jname = SourceNameMapper.getMethodName(spname, iename, function);
               if (jname != null)
               {
                  // found it
                  psuper = pp.phandle;
                  break;
               }
            }
         }
      }
      
      if (psuper == null)
      {
         String err = null;
         int errId = -1;
         if (function)
         {
            final String errMsg = 
               "SUPER version of user-defined function '%s' invoked but could not be found";
            errId = 6454;
            err = String.format(errMsg, iename);

            ErrorManager.recordOrShowError(errId, err, false, false, false);
         }
         else
         {
            character pname = searchSelf ? ppThis.pname : ppTarget.pname;

            errId = 6439;
            if (searchSelf)
            {
               final String errMsg = 
                  "Procedure %s %s has no SUPER procedure with internal procedure %s";
               err = String.format(errMsg, iename, pname.toStringMessage(), iename);
            }
            else
            {
               final String errMsg = 
                 "Procedure %s has no SUPER procedure with internal procedure %s";
               err = String.format(errMsg, pname.toStringMessage(), iename);
            }

            ErrorManager.recordOrThrowError(errId, err, false);
         }
      }
      
      return psuper;
   }
   
   /**
    * Get the THIS-PROCEDURE referent.
    * 
    * @return   See above.
    */
   static Object _thisProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.thisProcedures.peek();
      
      return referent;
   }

   /**
    * Check if the given referent is for a persistent procedure. This 
    * implements the PERSISTENT attribute for a procedure handle.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @return   <code>true</code> if the referent is a persistent procedure.
    */
   static logical isPersistent(Object referent)
   {
      return new logical(_isPersistent(referent));
   }

   /**
    * Check if the given referent is for a persistent procedure. This 
    * implements the PERSISTENT attribute for a procedure handle.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @return   <code>true</code> if the referent is a persistent procedure.
    */
   static boolean _isPersistent(Object referent)
   {
      return work.get().persistentProcedures.contains(referent);
   }
   
   /**
    * Check if the given referent is the top-level procedure (the startup procedure).
    * 
    * @param    referent
    *           The instance to be verified.
    *           
    * @return   <code>true</code> if the referent is the startup procedure.
    */
   static boolean isTopLevel(Object referent)
   {
      Deque<Object> procs = work.get().thisProcedures;
      return procs.getLast() == referent;
   }

   /**
    * Check if the given handle is a proxy procedure.
    * 
    * @param    h
    *           The handle to check.
    *           
    * @return   <code>true</code> if the handle is a proxy procedure.
    */
   static boolean isProxy(handle h)
   {
      return h != null && h._isValid() && (h.getResource() instanceof ProxyProcedureWrapper);
   }

   /**
    * Get the <code>SUPER-PROCEDURES</code> attribute of the specified 
    * external procedure instance.
    *
    * @param    referent
    *           An instance of an converted external procedure class.
    * 
    * @return   See above.
    */
   static character superProcedures(Object referent)
   {
      WorkArea wa = work.get();
      return superProceduresImpl(wa.getSuperProcedures(referent).values());
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h)
   {
      return addSuperProcedure(referent, h, (SearchMode) null);
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent,
    * using the explicit {@link SearchMode mode}.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h,SearchMode mode)
   {
      return addSuperProcedureImpl(h, mode, referent);
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent,
    * using the search-mode set as a string (one of the 
    * <code>SEARCH-TARGET</code> or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h, int mode)
   {
      return addSuperProcedure(referent, h, getSearchMode(mode));
   }

   /**
    * Remove the given handle from the specified referent's list of 
    * super-procedures.
    *  
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical removeSuperProcedure(Object referent, handle h)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }

      WorkArea wa = work.get();
      removeProcedureImpl(h, wa.getSuperProcedures(referent).keySet());

      // remove this procedure from the available list of super-procedures
      ProcedureData pp = locateProcedure(referent);
      for (Deque<Object> queue : pp.availSuperProcedures)
      {
         queue.remove(h.get());
      }
      
      return new logical(true);
   }
   
   /**
    * Get the <code>NEXT-SIBLING</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static handle getNextSibling(Object referent)
   {
      WorkArea wa = work.get();
      
      Object prev = null;
      Object next = null;
      for (Object pp : wa.persistentProcedures)
      {
         if (prev != null && prev == referent)
         {
            next = pp;
            break;
         }

         prev = pp;
      }
      
      if (next == null)
      {
         return new handle();
      }

      return locateProcedure(next).phandle;
   }
   
   /**
    * Get the <code>PREV-SIBLING</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static handle getPrevSibling(Object referent)
   {
      WorkArea wa = work.get();
      
      Object prev = null;
      for (Object pp : wa.persistentProcedures)
      {
         if (pp == referent)
         {
            break;
         }
         prev = pp;
      }
      
      if (prev == null)
      {
         return new handle();
      }

      return locateProcedure(prev).phandle;
   }

   static void setPrivateData(Object referent, character data)
   {
      ProcedureData pp = locateProcedure(referent);
      pp.privateData = new character(data);
   }

   /**
    * Get the <code>PRIVATE-DATA</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static character getPrivateData(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);

      return pp.privateData == null ? new character() 
                                    : new character(pp.privateData);
   }
   
   /**
    * Get the value of the ADM-DATA attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * 
    * @return   See above.
    */
   static character getADMData(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);
      return new character(pp.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute for the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @param    value
    *           The new value.
    */
   static void setADMData(Object referent, String value)
   {
      ProcedureData pp = locateProcedure(referent);
      pp.admData.assign(value);
   }
   
   /**
    * Gets the the unique ID number associated to the referent. Not the 
    * same as the handle value itself.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @return   The integer number of the children.
    */
   static integer getUniqueID(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);
      return new integer(pp.uniqueID);
   }

   /**
    * Get a handle list of super-procedures for the given referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static List<handle> getSuperProcedures(Object referent)
   {
      WorkArea wa = work.get();
      
      List<handle> superProcs = new LinkedList<handle>();
      // need to emit them in reverse order (they are kept in LIFO)
      for (ProcedureData pp : wa.getSuperProcedures(referent).values())
      {
         superProcs.add(0, pp.phandle);
      }
      
      return superProcs;
   }
   
   /**
    * Get a handle list of super-procedures for the session handle.
    *           
    * @return   See above.
    */
   static List<handle> getSuperProcedures()
   {
      return getSuperProcedures(null);
   }
   
   /**
    * Check if the given referent is for a system handle (i.e. for one of the
    * THIS-PROCEDURE, SOURCE-PROCEDURE or TARGET-PROCEDURE handles).
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static boolean systemHandle(Object referent)
   {
      WorkArea wa = work.get();
      return wa.thisProcedures.peek() == referent   ||
             wa.targetProcedures.peek() == referent ||
             wa.sourceProcedures.peek() == referent;
   }
   
   /**
    * Register the given {@link Scopeable} to be processed when the persistent procedure gets
    * deleted.
    * 
    * @param    referent
    *           The external program.
    * @param    target
    *           The {@link Scopeable} instance which needs delayed notification.
    */
   static void addScopeable(Object referent, Scopeable target)
   {
      ProcedureData pdata = locateProcedure(referent);
      pdata.scopeables.add(target);
   }
   
   /**
    * Register the given {@link Finalizable} to be processed when the persistent procedure gets
    * deleted.
    * 
    * @param    referent
    *           The external program.
    * @param    target
    *           The {@link Finalizable} instance which needs delayed notification.
    */
   static void addFinalizable(Object referent, Finalizable target)
   {
      ProcedureData pdata = locateProcedure(referent);
      pdata.finalizables.add(target);
   }
   
   /**
    * Gets the INSTANTIATING-PROCEDURE associated to the referent. 
    * 
    * @param    referent
    *           The external program.
    * 
    * @return   See above.
    */
   static handle getInstantiatingProcedure(Object referent)
   {
      ProcedureData pdata = locateProcedure(referent);
      return new handle(pdata.instProc);
   }
   
   /**
    * Resolve the instantiating procedure from the given handle: if the handle contains a valid
    * procedure, return a copy of it; else, return an unknown handle.
    * 
    * @param    hproc
    *           The handle instance representing an instantiating procedure.
    * 
    * @return   See above.
    */
   static handle resolveInstantiatingProcedure(handle hproc)
   {
      return hproc._isValid() ? new handle(hproc) : new handle();
   }
   
   /**
    * Compute the THIS-PROCEDURE handle using {@link Block} instance and the 
    * <code>this$</code> fields added by the compiler for each instance inner
    * class. When <code>this$0</code> is reached, it returns the handle for
    * that instance.
    *   
    * @return   See above.
    */
   private static handle thisProcedureImpl()
   {
      // this is the current executing procedure. need to determine the first
      // Block instance in the stack and after that resolve this$0 instance
      BlockDefinition block = TransactionManager.nearestExecutingBlock();

      // in hand written-code, we can end up with executing blocks being used
      // in static methods, called outside a real external procedure instance
      // in such cases, the Block instance has no this$ field; in such cases,
      // it is assumed that the previous THIS-PROCEDURE is the correct one,
      // so we return the top of the thisProcedures stack
      Field[] fields = block.block.getClass().getDeclaredFields();
      boolean foundThis = false;
      for (int i = 0; i < fields.length; i++)
      {
         Field f = fields[i];
         String fname = f.getName();
         if (fname.startsWith("this$"))
         {
            foundThis = true;
            break;
         }
      }
      Object outer = null;
      if (!foundThis)
      {
         outer = work.get().thisProcedures.peek();
      }
      else
      {
         outer = findRootEnclosingInstance(block.block);
         if (outer == null)
         {
            String className = block.block.getClass().getName();
            String ltClass = "com.goldencode.p2j.ui.LogicalTerminal$";
            if (className.startsWith(ltClass))
            {
               // if this is a persistent trigger invocation, then inherit
               // this-procedure from the previous block
               outer = work.get().thisProcedures.peek();
            }
            else
            {
               throw new RuntimeException("Could not determine THIS-PROCEDURE");
            }
         }
      }
      
      handle phandle = new handle();
      
      ExternalProgramWrapper extProg = new ExternalProgramWrapper(outer);
      phandle.set(extProg);
      
      return phandle;
   }

   /**
    * Find the root enclosing instance of the specified inner class instance.
    * It gets computed using the <code>this$</code> fields inserted by the 
    * compiler; it returns the instance refered by the <code>this$0</code> 
    * fields, when reached.
    * 
    * @param    instance
    *           The inner class instance.
    *           
    * @return   The root instance.
    */
   private static Object findRootEnclosingInstance(Object instance)
   {
      Field[] fields = instance.getClass().getDeclaredFields();
      for (int i = 0; i < fields.length; i++)
      {
         Field f = fields[i];
         String fname = f.getName();
         if (fname.startsWith("this$"))
         {
            try
            {
               f.setAccessible(true);
               Object outer = f.get(instance);
               
               return 
                  fname.equals("this$0") ? outer 
                                         : findRootEnclosingInstance(outer);
            }
            catch (Exception e)
            {
               e.printStackTrace();
               return null;
            }
         }
      }
      
      return null;
   }

   /**
    * Implementation for the SUPER-PROCEDURES attribute.
    *  
    * @param    superProcs
    *           A list of {@link ProcedureData} objects, in the order in 
    *           which they were added.
    *           
    * @return   The string representation of each super-procedure, separated
    *           by a comma. The last added super-procedure must be first in
    *           this list, the second-to-last must be the second in the list,
    *           etc.
    */
   private static character superProceduresImpl(Collection<ProcedureData> superProcs)
   {
      character sp = new character("");
      boolean first = true;
      ArrayDeque<ProcedureData> queue = new ArrayDeque<ProcedureData>();
      queue.addAll(superProcs);
      // need to emit them in a LIFO order
      while (!queue.isEmpty())
      {
         ProcedureData pp = queue.removeLast();
         int hashCode = pp.phandle.hashCode();
         sp.assign(TextOps.concat(sp, (!first ? "," : ""), hashCode));
         first = false;
      }
      
      return sp;
   }
   
   /**
    * The implementation of the ADD-SUPER-PROCEDURE method. When the
    * <code>referent</code> is null, this call is for the SESSION handle.
    * 
    * @param    h
    *           A handle to be added as super-procedure.
    * @param    mode
    *           The {@link SearchMode} for the passed handle.
    * @param    referent
    *           When <code>null</code> add to the SESSION:SUPER-PROCEDURES
    *           list. Else, add to the procedure associated with this 
    *           referent.
    * 
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   private static logical addSuperProcedureImpl(handle h, SearchMode mode, Object referent)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }
      
      WorkArea wa = work.get();
      Map<Object, ProcedureData> superProcs = wa.getSuperProcedures(referent);
      ProcedureData pData = locateProcedure(h.get());

      // the super-procedure to be added to the chain will always exist (either as the current
      // this-procedure or as a persistent procedure). the remove call is needed just to eliminate
      // it from the referent's super procedure chain, if was previously added 
      removeProcedureImpl(h, superProcs.keySet());

      
      if (!(mode == null || pData.searchMode == null || pData.searchMode.equals(mode)))
      {
         // if the mode is changed, show warning
         int errNum = 8914;
         String msg = "**Changing proc-search for procedure %s from  '%s' to '%s'";

         String progName = pData.pname.toStringMessage();
         String oldMode  = pData.searchMode.toString();
         String newMode  = mode.toString();
         
         msg = String.format(msg, progName, oldMode, newMode);

         // this is just a warning, the mode will be changed
         ErrorManager.recordOrShowError(errNum, msg, false, false, false);
      }

      pData.searchMode = mode;
      superProcs.put(h.get(), pData);

      return new logical(true);
   }

   /**
    * Removes the given procedure handle from the given procedure instances list.
    * 
    * @param    h
    *           The procedure handle to be removed.
    * @param    ppList
    *           The collection from which the procedure must be removed.
    */
   private static void removeProcedureImpl(handle h, Collection<Object> ppList)
   {
      WorkArea wa = work.get();
      
      Object hreferent = h.get();
      
      Iterator<Object> iter = ppList.iterator();
      while (iter.hasNext())
      {
         Object referent = iter.next();
         if (hreferent == referent)
         {
            iter.remove();
         }
      }

      wa.cleanup(hreferent);
   }
   
   /**
    * Get the {@link SearchMode} constant associated with the given mode, used
    * by ADD-SUPER-PROCEDURE calls. If the mode is not one of the SEARCH-SELF 
    * or SEARCH-TARGET values, or an {@link IllegalArgumentException} will be 
    * thrown.
    * <p>
    * The given <code>mode</code> is not compatible with the mode returned by
    * {@link SearchMode#toString()}.
    * 
    * @param    mode
    *           The string representation of the search-mode.
    *           
    * @return   For SEARCH-TARGET value, {@link SearchMode#SEARCH_TARGET} 
    *           will be returned. For SEARCH-SELF value, 
    *           {@link SearchMode#SEARCH_SELF} will be returned.
    *
    * @throws   IllegalArgumentException
    *           If the mode is not of a known value.
    */
   private static SearchMode getSearchMode(int mode)
   {
      if (mode == SEARCH_SELF)
         return SearchMode.SEARCH_SELF;
      
      if (mode == SEARCH_TARGET)
         return SearchMode.SEARCH_TARGET;
      
      // if not a known mode, default to "search mode not specified"
      return null;
   }
   
   /**
    * Locate the {@link ProcedureData} instance associated with the given external program
    * instance.
    * <p>
    * This will find any persistent procedure or any procedure active on the stack.
    * 
    * @param    referent
    *           The external program instance.
    * 
    * @return   See above.
    */
   private static ProcedureData locateProcedure(Object referent)
   {
      return work.get().pMap.get(referent);
   }
   
   /**
    * Delete the given resources, as long as they are still valid.
    * 
    * @param    resources
    *           The resources to be deleted.
    */
   private static void deleteResources(Set<WrappedResource> resources)
   {
      for (WrappedResource target : resources)
      {
         if (target.valid())
         {
            ((Deletable) target).delete();
         }
      }
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   implements Scopeable
   {
      /**
       * The list of all registered persistent procedures, via RUN PERSISTENT
       * calls.
       */
      private Deque<Object> persistentProcedures = new ArrayDeque<Object>();
      
      /**
       * A map containing the list of super-procedures, for the session,
       * by each instance. 
       */
      private Map<Object, ProcedureData> sessionSuperProcs = 
         new LinkedHashMap<Object, ProcedureData>();
      
      /** 
       * This map holds the attributes for all registered external procedures, by instance.
       * It includes all persistent procedures and all procedures active on the stack.
       */
      private Map<Object, ProcedureData> pMap = new HashMap<Object, ProcedureData>();
      
      /** Stack of target-procedures, by instance. */
      private Deque<Object> targetProcedures = new ArrayDeque<Object>();

      /** Stack of source-procedures, by instance. */
      private Deque<Object> sourceProcedures = new ArrayDeque<Object>();
      
      /** Stack of this-procedures, by instance. */
      private Deque<Object> thisProcedures = new ArrayDeque<Object>();
      
      /** Resources pending to registered with THIS-PROCEDURE when the next block is entered. */
      private Set<WrappedResource> pendingResources = new HashSet<>();
      
      /**
       * Indicates the current external program when a scope or finalizable notification occurs.
       */
      private Object processedProcedure = null;
     
      /**
       * Handle scope notifications, to maintain proper source, target, and
       * this-procedure references.
       */
      @Override
      public void scopeStart()
      {
         BlockType bt = TransactionManager.getBlockType();

         if (bt.equals(BlockType.EXTERNAL_PROC) ||
             bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION)      ||
             bt.equals(BlockType.TRIGGER))
         {
            // push this-procedure to the stack and populate other data
            handle hthis = thisProcedureImpl();
            Object referent = hthis.get();
            ProcedureData pdata = pMap.get(referent);

            if (pdata == null)
            {
               pdata = new ProcedureData(hthis, null);
               pMap.put(referent, pdata);
            }
            
            // setting INSTANTIATING-PROCEDURE for external procedure block can be done only 
            // after ProcedureData is created
            if (bt.equals(BlockType.EXTERNAL_PROC) && TransactionManager.getNestingLevel() > 2)
            {
               // if this is not the entry point, set the INSTANTIATING-PROCEDURE as the current 
               // THIS-PROCEDURE
               pdata.instProc.assign(ProcedureManager.thisProcedure());
            }

            thisProcedures.push(referent);
            
            if (bt.equals(BlockType.EXTERNAL_PROC))
            {
               if (TransactionManager.getNestingLevel() == 2 && !SessionUtils._isRemote())
               {
                  // if this is the entry-point and not a remote proc, set it as target procedure
                  pushTargetProcedure(hthis);
               }
            }
            
            WidgetPool.scopeStart();

            pdata.staticResources.push(new HashSet<WrappedResource>());
         }

         if (!TransactionManager.isGlobalBlock())
         {
            ProcedureData pdata = locateProcedure(_thisProcedure());
   
            // process the newly created resources
            for (WrappedResource res : pendingResources)
            {
               if (res instanceof HandleResource)
               {
                  ((HandleResource) res).setInstantiatingProcedure(new handle(pdata.phandle));
               }
            }

            // collect the static resources in the correct scope
            pdata.staticResources.peek().addAll(pendingResources);
            pendingResources.clear();
         }
      }

      /**
       * Handle scope notifications, to maintain proper source, target, and
       * this-procedure references.
       */
      @Override
      public void scopeFinished()
      {
         BlockType bt = TransactionManager.getBlockType();
         if (bt.equals(BlockType.EXTERNAL_PROC) ||
             bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION)      ||
             bt.equals(BlockType.TRIGGER))
         {
            // pop this-procedure from the stack and cleanup
            Object referent = thisProcedures.pop();
            ProcedureData pdata = locateProcedure(referent);
            
            // notify the widget-pool management a procedure has finished; this needs to be done
            // here to ensure that the correct THIS-PROCEDURE handle is available
            boolean extProc = bt.equals(BlockType.EXTERNAL_PROC);
            boolean persistent = _isPersistent(referent);

            // if we are terminating the entry-point procedure, clean the target-procedure too
            if (TransactionManager.getNestingLevel() == 2)
            {
               popTargetProcedure();
            }
            cleanup(referent);
            
            if (!(extProc && persistent))
            {
               // delete the static resources created in an internal proc/function/trigger or
               // if the external program is not persistent
               deleteResources(pdata.staticResources.pop());
            }

            WidgetPool.scopeFinished(referent, extProc, persistent);
         }
      }
      
      /**
       * Provides a notification that an external scope is about to be deleted.
       * <p>
       * This implementation is a no-op.
       */
      @Override
      public void scopeDeleted()
      {
         // no-op
      }
      
      /**
       * Get the map of session's super-procedures.
       */
      public Map<Object, ProcedureData> getSessionSuperProcedures()
      {
         return getSuperProcedures(null);
      }

      /**
       * Get the map of the super-procedures for the given referent.
       * 
       * @param    referent
       *           The external program instance for which the super-procedures
       *           are needed.
       */
      public Map<Object, ProcedureData> getSuperProcedures(Object referent)
      {
         return referent == null ? sessionSuperProcs 
                                 : pMap.get(referent).superProcedures;
      }
      
      /**
       * Attempt cleanup of the given external program instance. If is no
       * longer used as a persistent procedure, super procedure, or as a 
       * system handle, the referent will be removed from the {@link #pMap}
       * map.
       * 
       * @param    referent
       *           The external program instance for which needs to be cleaned
       *           up.
       */
      public void cleanup(Object referent)
      {
         // if no longer used, remove from global map
         if (!thisProcedures.contains(referent)                   &&
             !targetProcedures.contains(referent)                 &&
             !sourceProcedures.contains(referent)                 &&
             !persistentProcedures.contains(referent)             &&
             !sessionSuperProcs.containsKey(referent)             &&
             !superProcForExtProc(referent, persistentProcedures) &&
             !superProcForExtProc(referent, sessionSuperProcs.keySet()))
         {
            pMap.remove(referent);
         }
      }
      
      /**
       * Check if the given referent is a super-procedure of one of the
       * procedures in the given list.
       * 
       * @param    referent
       *           The external program instance.
       * @param    progs
       *           The list of external programs for which their 
       *           super-procedures list will be searched for the referent.
       *            
       * @return   <code>true</code> if it was found as a super-procedure for
       *           one of the given procedures.
       */
      private boolean superProcForExtProc(Object referent, 
                                          Collection<Object> progs)
      {
         for (Object prog : progs)
         {
            ProcedureData pp = pMap.get(prog);
            for (Object sreferent : pp.superProcedures.keySet())
            {
               if (sreferent == referent)
               {
                  return true;
               }
            }
         }
         
         return false;
      }
   }

   /**
    * Contains data about a subscription made by a certain procedure.
    */
   private static class Subscription
   {
      /** The subscriber. */
      private final Object subscriber;
      
      /** The subscribed event. */
      private final character event = new character();
      
      /** The method to be invoked. May be <code>null</code>. */
      private final character method;
      
      /** The publisher to match. <code>null</code> represents any source. */
      private final Object publisher;
      
      /** Flag indicating that this subscription can't be removed. */
      public boolean readOnly = false;
      
      /**
       * Build a new subscription with the specified data.
       * 
       * @param    subscriber
       *           The subscriber procedure.
       * @param    event
       *           The event to subscribe.
       * @param    publisher
       *           The event's publisher. May be <code>null</code>.
       * @param    method
       *           The method to be invoked. When <code>null</code>, the event name will be used
       *           as the method name.
       */
      public Subscription(Object subscriber, character event, Object publisher, character method)
      {
         this.subscriber = subscriber;
         this.publisher = publisher;
         this.event.assign(event);
         this.method = new character(method == null ? event : method);
      }
   }

   /**
    * Container for various procedure data.
    */
   private static class ProcedureData
   {
      /** The subscriptions to which this procedure is registered. */
      private final List<Subscription> subscriptions = new ArrayList<Subscription>();
      
      /** The list of super-procedures for this procedure, by external program instance. s*/
      private final Map<Object, ProcedureData> superProcedures = 
         new LinkedHashMap<Object, ProcedureData>();
      
      /** 
       * The map of name of functions defined IN handle in this external program, to their 
       * {@link Resolvable} instance. 
       */
      private final Map<String, Resolvable> inHandleFunctions = new HashMap<String, Resolvable>();
      
      /** Available super-procedures for this procedure handle, for each call. */
      private final Deque<Deque<Object>> availSuperProcedures = new ArrayDeque<Deque<Object>>();
      
      /**
       * A handle instance for this procedure, so that other handles can be instantiated from it.
       */
      private handle phandle = null;
      
      /** The current search-mode for this external program. */
      private SearchMode searchMode = null;
      
      /** The legacy 4GL name of this external program. */
      private character pname = null;
      
      /** The private data of this procedure handle. */
      private character privateData = null;
      
      /** Implementation of ADM-DATA attribute. */
      private final character admData = new character();

      /** Implementation of UNIQUE-ID attribute. */
      private final integer uniqueID = new integer();
      
      /** Handle referencing the instantiating procedure. */
      private final handle instProc = new handle();
      
      /**
       * Set of {@link Scopeable} instances which need notification when the persistent procedure 
       * gets deleted.
       */
      private Set<Scopeable> scopeables = new HashSet<>();

      /**
       * Set of {@link Finalizable} instances which need notification when the persistent procedure 
       * gets deleted.
       */
      private Set<Finalizable> finalizables = new HashSet<>();

      /**
       * A stack holding the resources created in each top-level block.
       */
      private Deque<Set<WrappedResource>> staticResources = new ArrayDeque<>();
      
      /**
       * Create a new {@link ProcedureData} instance using the given 
       * procedure handle and search-mode.
       * 
       * @param    phandle
       *           The procedure handle.
       * @param    searchMode
       *           The search-mode. May be null.
       */
      ProcedureData(handle phandle, SearchMode searchMode)
      {
         this.phandle = new handle(phandle);
         this.searchMode = searchMode;
         // at this point, we know the phandle references a valid procedure, 
         // so the name can be retrieved safely.
         this.pname = getFileName(phandle.get());
         this.uniqueID.assign(new integer(UniqueIdGenerator.getUniqueId(IdKind.PROCEDURE)));
      }
      
      /**
       * Check if the search-mode for this external program is SEARCH-SELF.
       * 
       * @return   See above.
       */
      boolean isSearchSelf()
      {
         return searchMode == null || searchMode == SearchMode.SEARCH_SELF;
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Initializes the work area, the first time it is requested within a
       * new context. Registers the {@link WorkArea} instance with the
       * {@link TransactionManager}, for scope notifications.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         if (!ErrorManager.isHeadless())
         {
            TransactionManager.registerScopeable(wa);
         }

         return wa;
      }
   }
}